@import url("resource:///org/gnome/builder/theme/Adwaita-shared.css");

/*
 * Project Tree
 */
GbProjectTree.view {
  color: #eeeeec;
  background-color: mix(#292929, #2e3436, 0.8);
}
GbProjectTree.view:backdrop {
  color: #a1a1a1;
  background-color: #2e3436;
}
GbProjectTree.view.cell:selected {
  color: #fff;
  background-color: #215d9c;
}

