/* ide-clang-highlighter.h
 *
 * Copyright (C) 2015 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef IDE_CLANG_HIGHLIGHTER_H
#define IDE_CLANG_HIGHLIGHTER_H

#include "ide-highlighter.h"

G_BEGIN_DECLS

#define IDE_TYPE_CLANG_HIGHLIGHTER (ide_clang_highlighter_get_type())

#define IDE_CLANG_HIGHLIGHTER_TYPE          "c:type"
#define IDE_CLANG_HIGHLIGHTER_FUNCTION_NAME "c:function-name"
#define IDE_CLANG_HIGHLIGHTER_ENUM_NAME     "c:enum-name"
#define IDE_CLANG_HIGHLIGHTER_MACRO_NAME    "c:macro-name"

G_DECLARE_FINAL_TYPE (IdeClangHighlighter, ide_clang_highlighter,
                      IDE, CLANG_HIGHLIGHTER, IdeHighlighter)

G_END_DECLS

#endif /* IDE_CLANG_HIGHLIGHTER_H */
