@import url("resource:///org/gnome/builder/theme/Adwaita-shared.css");

/*
 * Project Tree
 */
GbProjectTree.view {
  color: #555753;
  background-color: #eeeeec;
}
GbProjectTree.view:backdrop {
  color: #888a85;
  background-color: #efefef;
}
GbProjectTree.view.cell:selected {
  color: #fff;
  background-color: #4a90d9;
}
