
/* Generated data (by glib-mkenums) */

#ifndef __IDE_ENUMS_H__
#define __IDE_ENUMS_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "libide/ide-buffer.h" */
GType ide_buffer_line_flags_get_type (void);
#define IDE_TYPE_BUFFER_LINE_FLAGS (ide_buffer_line_flags_get_type ())

/* enumerations from "libide/ide-diagnostic.h" */
GType ide_diagnostic_severity_get_type (void);
#define IDE_TYPE_DIAGNOSTIC_SEVERITY (ide_diagnostic_severity_get_type ())

/* enumerations from "libide/ide-highlighter.h" */
GType ide_highlight_result_get_type (void);
#define IDE_TYPE_HIGHLIGHT_RESULT (ide_highlight_result_get_type ())

/* enumerations from "libide/ide-source-view.h" */
GType ide_source_view_mode_type_get_type (void);
#define IDE_TYPE_SOURCE_VIEW_MODE_TYPE (ide_source_view_mode_type_get_type ())
GType ide_source_view_theatric_get_type (void);
#define IDE_TYPE_SOURCE_VIEW_THEATRIC (ide_source_view_theatric_get_type ())
GType ide_source_view_movement_get_type (void);
#define IDE_TYPE_SOURCE_VIEW_MOVEMENT (ide_source_view_movement_get_type ())
G_END_DECLS

#endif /* __IDE_ENUMS_H__ */

/* Generated data ends here */

